/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package ZEN_IDL;

import java.io.*;
import java.util.*;

import ZEN_IDL.ast.*;
import ZEN_IDL.be.*;
import util.*;

/*********************************************************
* Main class for running the IDL compiler
*
* @author Olav Haugan
* @author David McKinnon
*********************************************************/
public final class IdlCompiler
{
  
  /*********************************************************
  * main function. Reads command line arguments and
  * calls the idl compiler.
  *
  * @param String[] The list of command line arguments
  *********************************************************/
  public static void main( String args[] )
  {
    if( args.length < 2 )
    {
      System.out.println( "Usage: IdlCompiler <IDL File> <MQC config file>" );
      return;
    }  

    FileInputStream idlFile;
    try
    {
      idlFile = new FileInputStream( new File(args[0]) );
    }
    catch( FileNotFoundException fnf )
    {
      System.out.println( "File not found: " + args[0]);
      return;
    }

    IDLParser parser = new IDLParser( idlFile );
    SimpleNode parseTree;
  
    //Parse the IDL file
    try
    {
      parseTree = parser.specification();
    }
    catch( ParseException pe )
    {
      System.out.println( "Error parsing IDL file." );
      pe.printStackTrace(); 
      return;
    }

    //Create the symbol table
    IDLParserVisitor vSymbolTable = new SymbolTableVisitor();
    VisitorState mState = new VisitorState();
    parseTree.jjtAccept(vSymbolTable , new VisitorState() );
    if( mState.getErrorOccured() ){
      System.exit(0);
    }

    //Check the IDL constraints in the symbol table
    MQCConfig config = new MQCConfig(args[1]);
    MQCConstraintCheckerVisitor checkerVisitor = new MQCConstraintCheckerVisitor(config);
    parseTree.jjtAccept( checkerVisitor , new VisitorState() );
        
    //Generate the code
    //IDLParserVisitor codeGeneratorVisitor = new MQCVisitor(config);
    MQCVisitor codeGeneratorVisitor = new MQCVisitor(config);
    parseTree.jjtAccept( codeGeneratorVisitor , new VisitorState());

    // Get the package names which are needed by the Makefile
    String packageNames = codeGeneratorVisitor.getPackageNames();

    //Generate config file
    MQCConfigGenerator configGen = new MQCConfigGenerator(config);
    configGen.generate();
    
    //Generate the Config_macros file
    MQCConfigMacrosGenerator configMacrosGen = new MQCConfigMacrosGenerator(config);
    configMacrosGen.generate();

    //Generate the Makefile
    MQCMakefileGen makefile = new MQCMakefileGen(config, packageNames);
    makefile.generate();
  }
}
