/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package ZEN_IDL.be;

import ZEN_IDL.ast.*;
import ZEN_IDL.be.symbolTableUtils.*;
import ZEN_IDL.be.helperUtils.*;
import util.*;

import java.io.*;
import java.util.*;

/*********************************************************
* Generates stubs and skeleton inline marshalling and 
* unmarshalling code for use with MQCIOP protocol.
*
* @author Olav Haugan
*
* 16.10.01
*********************************************************/
public class MQCInline_MQCIOP_MarshalHelper extends MQCAbstractMarshalHelper
{
	private boolean floatDef; /* Float variable declared or not */
	private boolean doubleDef; /* Double variable declared or not */
	private	boolean longDef; /* long variables declared or not */			

	/*********************************************************
	* Constructor. Initializes variables.
	* 
	* @param Hashtable A method mapping object
	*        
	*********************************************************/
	public MQCInline_MQCIOP_MarshalHelper(MQCConfig config )
	{
		super(config);
		mqcIOP = true;
	}

	/*********************************************************
	* Initializes the marshaller.
	* These variables are used to keep track of declared
	* variables so that we don't declare the same variable
	* twice.
	*********************************************************/
	public void init()
	{
		floatDef = false;
		doubleDef = false;
		longDef = false;		
	}

	/*********************************************************
	* Prints the marshalling code for 
	* either the client or server. 
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethod The method
	*
	* @param boolean true if code is for client, false
  *                if for server.
	*
	* @return int The size of buffer needed.
	*********************************************************/
	public int printMarshal( MQCPrinter pw, InterfaceMethod iMethod, boolean Client )
	{
		if(Client)
			return printClientMarshal(pw, iMethod );
		else
			return printServerMarshal(pw, iMethod );
	}

	/*********************************************************
	* Prints the marshalling code for 
	* either the client or server. 
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethod The method
	*
	* @param boolean true if code is for client, false
  *                if for server.
	*
	*********************************************************/
	public void printUnMarshal( MQCPrinter pw, InterfaceMethod iMethod, boolean Client )
	{
		if(Client)
			printClientUnMarshal(pw, iMethod );
		else
			printServerUnMarshal(pw, iMethod );
	}
	
	/*********************************************************
	* Prints the inline client marshalling code
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethod The method
	*
	*********************************************************/
	public int printClientMarshal(MQCPrinter pw, InterfaceMethod iMethod )
	{
		Enumeration e = iMethod.paramVector.elements();
		boolean padding = config.isPadding();
		int index = Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REQ_HDR;
		if(config.isCompressing())
		{
			int funNo = MQCCompress.compressMethod(iMethod.name);
			index += String.valueOf(funNo).length() + 1;
		}
		else
		{
			index += iMethod.name.length() + 1;
		}	
				
		e = iMethod.paramVector.elements();									
		InterfaceMethodParam arg = null;
		
		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			if( !arg.paramInOut.equals("out") )
			{
				if(codeGenHelper.isPrimitiveType(arg.paramType))
				{
					if(padding)
					{
						index = codeGenHelper.align(arg.paramType, index);
					}
					if(arg.paramType.equals("float"))
						index = doMarshalFloat(pw, arg, index);
					else if(arg.paramType.equals("double"))
						index = doMarshalDouble(pw, arg, index);
					else if(arg.paramType.equals("boolean"))
						index = doMarshalBoolean(pw, arg, index);
					else
						index = doMarshal(pw, arg, index);
				}
			}
		}
		return index;
	}


	/*********************************************************
	* Prints the inline marshalling code for basic
	* types.
	*
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethodParam The method parameter
	
	* @param int The current index
	
	* @return int The new index
	*********************************************************/
	public int doMarshal(MQCPrinter pw, InterfaceMethodParam arg, int index )
	{
		int paramSize = codeGenHelper.getParamSize(arg.paramType);
		
		for(int i = paramSize-1; i > 0; i--)
		{
			pw.print("_msg[" + index++ + "] = (byte) ");
			if(!arg.paramInOut.equals("in"))
				pw.nprintln("(" + arg.paramName + ".value >> " + (8 * i) + ");" );
			else
				pw.nprintln("(" + arg.paramName + " >> " + (8 * i) + ");" );
		}
		pw.print("_msg[" + index++ + "] = (byte) ");
		if(!arg.paramInOut.equals("in"))
			pw.nprintln(arg.paramName + ".value;" );
		else
			pw.nprintln(arg.paramName + ";" );
		return index;
	}	

	/*********************************************************
	* Prints the inline server marshalling code
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethod The method
	*
	*********************************************************/
	public int printServerMarshal(MQCPrinter pw, InterfaceMethod iMethod )
	{
		Enumeration e = iMethod.paramVector.elements();
		InterfaceMethodParam arg = null;
		boolean padding = config.isPadding();
		int index = Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR;

		int paramSize = codeGenHelper.getParamSize(iMethod.returnType);
		
		if(padding)
		{
			index = codeGenHelper.align(iMethod.returnType, index );
		}
		if(iMethod.returnType.equals("float"))
			index = doMarshalFloat(pw, null, index);
		else if(iMethod.returnType.equals("double"))
			index = doMarshalDouble(pw, null, index);
		else if(iMethod.returnType.equals("boolean"))
			index = doMarshalBoolean(pw, null, index);
		else if(!iMethod.returnType.equals("void"))
		{
			for(int i = paramSize-1; i > 0; i--)
			{
				pw.print("_msg[" + index++ + "] = (byte) ");
				pw.nprintln("(_result >> " + (8 * i) + ");" );
			}
			pw.println("_msg[" + index++ + "] = (byte) _result;");
		}
		e = iMethod.paramVector.elements();
		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			if( !arg.paramInOut.equals("in") )
			{
				if(padding)
				{
					index = codeGenHelper.align(arg.paramType, index );
				}
				if(codeGenHelper.isPrimitiveType(arg.paramType))
				{
					if(arg.paramType.equals("float"))
						index = doMarshalFloat(pw, arg, index);
					else if(arg.paramType.equals("double"))
						index = doMarshalDouble(pw, arg, index);
					else if(arg.paramType.equals("boolean"))
						index = doMarshalBoolean(pw, arg, index);
					else
						index = doMarshal(pw, arg, index);					
				}
			}
		}
		return index;
	}


	/*********************************************************
	* Prints the inline client unmarshalling code
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethod The method
	*
	*********************************************************/
	public void printClientUnMarshal(MQCPrinter pw, InterfaceMethod iMethod )
	{
		Enumeration e = iMethod.paramVector.elements();
		InterfaceMethodParam arg;
		boolean padding = config.isPadding();
		int index = Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR;
		int tempIndex;
				
		int paramSize;
		
		if(!iMethod.returnType.equals("void"))
		{
			if(codeGenHelper.isPrimitiveType(iMethod.returnType))
			{
				paramSize = codeGenHelper.getParamSize(iMethod.returnType);
				if(padding)
				{
					index = codeGenHelper.align(iMethod.returnType, index);
				}
				if(iMethod.returnType.equals("float"))
					index = doUnMarshalFloat(pw, null, index);
				else if(iMethod.returnType.equals("double"))
					index = doUnMarshalDouble(pw, null, index);
				else if(iMethod.returnType.equals("boolean"))
					index = doUnMarshalBoolean(pw, null, index );
				else if(iMethod.returnType.equals("long long") || iMethod.returnType.equals("unsigned long long"))
					index = doUnMarshalLong(pw, null, index );
				else
				{
					if(paramSize > 1)
					{
						pw.println("if(bigEndian)");
						pw.println("{");
						pw.setIndent(pw.getIndent()+1);
					}
					
					pw.print("_result = (" + codeGenHelper.mapIDL2Java( iMethod.returnType ) + ") (");
					tempIndex = index;
					for(int i = paramSize-1; i > 0; i--)
					{
						pw.nprint("((_msg[" + index++ + "] & 0xFF) << " + (8 * i) + ") | ");
					}
					pw.nprintln("(_msg[" + index++ + "] & 0xFF));");

					if(paramSize > 1)
					{
						pw.setIndent(pw.getIndent()-1);
						pw.println("}");
						pw.println("else");
						pw.println("{");
						pw.setIndent(pw.getIndent()+1);
					
						pw.print("_result = (" + codeGenHelper.mapIDL2Java( iMethod.returnType ) + ") (");
						for(int i = 0; i < paramSize-1; i++)
						{
							pw.nprint("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * i) + ") | ");
						}
						pw.nprintln("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * (paramSize-1)) + "));");
					
						pw.setIndent(pw.getIndent()-1);
						pw.println("}");
					}
				}
			}
		}
		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			if( !arg.paramInOut.equals("in") )
			{
				if(codeGenHelper.isPrimitiveType(arg.paramType))
				{
					paramSize = codeGenHelper.getParamSize(arg.paramType);
					if(padding)
					{
						index = codeGenHelper.align(arg.paramType, index);
					}
					if(arg.paramType.equals("float"))
						index = doUnMarshalFloat(pw, arg, index);
					else if(arg.paramType.equals("double"))
						index = doUnMarshalDouble(pw, arg, index);
					else if(arg.paramType.equals("boolean"))
						index = doUnMarshalBoolean(pw, arg, index);
					else if(arg.paramType.equals("long long") || arg.paramType.equals("unsigned long long"))
						index = doUnMarshalLong(pw, arg, index);
					else
						index = doUnMarshal(pw, arg, index);
				}
			}
		} //while
	}

	/*********************************************************
	* Prints the server inline unmarshalling code
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethod The method
	*
	*********************************************************/
	public void printServerUnMarshal(MQCPrinter pw, InterfaceMethod iMethod )
	{
		Enumeration e = iMethod.paramVector.elements();
		InterfaceMethodParam arg;
		boolean padding = config.isPadding();
		int index = Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REQ_HDR;
		if(config.isCompressing())
		{
			int funNo = MQCCompress.compressMethod(iMethod.name);
			index += String.valueOf(funNo).length() + 1;
		}
		else
		{
			index += iMethod.name.length() + 1;
		}	

		int paramSize = 0;
		
		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			if( !arg.paramInOut.equals("out") )
			{
				if(padding)
				{
					index = codeGenHelper.align(arg.paramType, index);
				}

				if(codeGenHelper.isPrimitiveType(arg.paramType))
				{
					if(!arg.paramInOut.equals("in"))
					{
						pw.print(codeGenHelper.mapIDL2JavaHolder( arg.paramType ) );
						pw.nprint( " " + arg.paramName + " = " );
						pw.nprintln("new " + codeGenHelper.mapIDL2JavaHolder( arg.paramType ) + "();" );
					}
					else
					{
						pw.print(codeGenHelper.mapIDL2Java( arg.paramType ) );
						pw.nprintln( " " + arg.paramName + ";" );
					}
					if(arg.paramType.equals("float"))
						index = doUnMarshalFloat(pw, arg, index);
					else if(arg.paramType.equals("double"))
						index = doUnMarshalDouble(pw, arg, index);
					else if(arg.paramType.equals("boolean"))
						index = doUnMarshalBoolean(pw, arg, index);
					else if(arg.paramType.equals("long long") || arg.paramType.equals("unsigned long long"))
						index = doUnMarshalLong(pw, arg, index);
					else
						index = doUnMarshal(pw, arg, index);
				}
			}
			else
			{
				pw.print(codeGenHelper.mapIDL2JavaHolder( arg.paramType ) );
				pw.nprint( " " + arg.paramName + " = " );				
				pw.nprintln("new " + codeGenHelper.mapIDL2JavaHolder( arg.paramType ) + "();" );
			}
		}
	}

	/*********************************************************
	* Prints the un-marshaling of basic types
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethodParam The method parameter
	*
	* @param int The current index
	*	
	* @return int The new index
	*********************************************************/
	public int doUnMarshal(MQCPrinter pw, InterfaceMethodParam arg, int index)
	{
		int paramSize = codeGenHelper.getParamSize(arg.paramType);
		int tempIndex;
		
		if(paramSize > 1)
		{
			pw.println("if(bigEndian)");
			pw.println("{");
			pw.setIndent(pw.getIndent()+1);
		}
		
		if(!arg.paramInOut.equals("in"))
			pw.print(arg.paramName + ".value = (" + codeGenHelper.mapIDL2Java( arg.paramType ) + ") (");
		else
			pw.print(arg.paramName + " = (" + codeGenHelper.mapIDL2Java( arg.paramType ) + ") (");

		tempIndex = index;
		for(int i = paramSize-1; i > 0; i--)
		{
			pw.nprint("((_msg[" + index++ + "] & 0xFF) << " + (8 * i) + ") | ");
		}
		pw.nprintln("(_msg[" + index++ + "] & 0xFF));");

		if(paramSize > 1)
		{
			pw.setIndent(pw.getIndent()-1);
			pw.println("}");
			pw.println("else");
			pw.println("{");
			pw.setIndent(pw.getIndent()+1);
			if(!arg.paramInOut.equals("in"))
				pw.print(arg.paramName + ".value = (" + codeGenHelper.mapIDL2Java( arg.paramType ) + ") (");
			else
				pw.print(arg.paramName + " = (" + codeGenHelper.mapIDL2Java( arg.paramType ) + ") (");
			for(int i = 0; i < paramSize-1; i++)
			{
				pw.nprint("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * i) + ") | ");
			}
			pw.nprintln("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * (paramSize-1)) + "));");
			pw.setIndent(pw.getIndent()-1);
			pw.println("}");
		}
		return index;
	}

	/*********************************************************
	* Prints the marshaling of boolean 
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethodParam The method parameter
	*
	* @param int The current index
	*	
	* @return int The new index
	*********************************************************/
	public int doMarshalBoolean(MQCPrinter pw, InterfaceMethodParam arg, int index)
	{
		if(arg == null)
		{
			pw.println("_msg[" + index++ + "] = (byte) (_result ? 1 : 0);");		
		}
		else
		{
			if(arg.paramInOut.equals("in"))
				pw.println("_msg[" + index++ + "] = (byte) (" + arg.paramName + " ? 1 : 0);");
			else
				pw.println("_msg[" + index++ + "] = (byte) (" + arg.paramName + ".value ? 1 : 0);");
		}
		return index;
	}
	
	/*********************************************************
	* Prints the marshaling of float 
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethodParam The method parameter
	*
	* @param int The current index
	*	
	* @return int The new index
	*********************************************************/
	public int doMarshalFloat(MQCPrinter pw, InterfaceMethodParam arg, int index)
	{
		if(!floatDef)
			pw.print("int fval = ");
		else
			pw.print("fval = ");
		floatDef = true;
		
		if(arg == null)
		{
			pw.nprintln("java.lang.Float.floatToIntBits(_result);");		
		}
		else
		{
			if(!arg.paramInOut.equals("in"))
				pw.nprintln("java.lang.Float.floatToIntBits(" + arg.paramName + ".value);");
			else
				pw.nprintln("java.lang.Float.floatToIntBits(" + arg.paramName + ");");		
		}
		int paramSize = codeGenHelper.getParamSize("float");
		for(int i = paramSize-1; i > 0; i--)
		{
			pw.print("_msg[" + index++ + "] = (byte) ");
			pw.nprintln("(fval >> " + (8 * i) + ");" );
		}
		pw.println("_msg[" + index++ + "] = (byte) fval;");
		return index;
	}


	/*********************************************************
	* Prints the marshaling of double
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethodParam The method parameter
	*
	* @param int The current index
	*	
	* @return int The new index
	*********************************************************/
	public int doMarshalDouble(MQCPrinter pw, InterfaceMethodParam arg, int index)
	{
		if(!doubleDef)
			pw.print("long dval = ");
		else
			pw.print("dval = ");
		doubleDef = true;
		
		if(arg == null)
		{
			pw.nprintln("java.lang.Double.doubleToLongBits(_result);");
		}
		else
		{
			if(!arg.paramInOut.equals("in"))
				pw.nprintln("java.lang.Double.doubleToLongBits(" + arg.paramName + ".value);");
			else
				pw.nprintln("java.lang.Double.doubleToLongBits(" + arg.paramName + ");");
		}

		int paramSize = codeGenHelper.getParamSize("double");
		for(int i = paramSize-1; i > 0; i--)
		{
			pw.print("_msg[" + index++ + "] = (byte) ");
			pw.nprintln("(dval >> " + (8 * i) + ");" );
		}
		pw.println("_msg[" + index++ + "] = (byte) dval;");
		return index;
	}

	/*********************************************************
	* Prints the un-marshaling of boolean 
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethodParam The method parameter
	*
	* @param int The current index
	*	
	* @return int The new index
	*********************************************************/
	public int doUnMarshalBoolean(MQCPrinter pw, InterfaceMethodParam arg, int index)
	{
		if(arg == null)
		{
			pw.println("_result = (_msg[" + index++ + "] == 1) ? true : false;");		
		}
		else
		{
			if(arg.paramInOut.equals("in"))
				pw.println(arg.paramName + " = (_msg[" + index++ + "] == 1) ? true : false;");
			else
				pw.println(arg.paramName + ".value = (_msg[" + index++ + "] == 1) ? true : false;");
		}
		return index;
	}

	/*********************************************************
	* Prints the un-marshaling of float
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethodParam The method parameter
	*
	* @param int The current index
	*	
	* @return int The new index
	*********************************************************/
	public int doUnMarshalFloat(MQCPrinter pw, InterfaceMethodParam arg, int index)
	{
		if(!floatDef)
			pw.println("int fval = 0;");
		floatDef = true;
		int tempIndex;
		
		int paramSize = codeGenHelper.getParamSize("float");
		pw.println("if(bigEndian)");
		pw.println("{");
		pw.setIndent(pw.getIndent()+1);

		pw.print("fval = (int) (" );
		tempIndex = index;
		for(int i = paramSize-1; i > 0; i--)
		{
			pw.nprint("((_msg[" + index++ + "] & 0xFF) << " + (8 * i) + ") | ");
		}
		pw.nprintln("(_msg[" + index++ + "] & 0xFF));");

		pw.setIndent(pw.getIndent()-1);
		pw.println("}");
		pw.println("else");
		pw.println("{");
		pw.setIndent(pw.getIndent()+1);

		pw.print("fval = (int) (" );
		for(int i = 0; i < paramSize-1; i++)
		{
			pw.nprint("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * i) + ") | ");
		}
		pw.nprintln("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * (paramSize-1)) + "));");

		pw.setIndent(pw.getIndent()-1);
		pw.println("}");

		if(arg == null)
		{
			pw.println("_result = java.lang.Float.intBitsToFloat(fval);");
		}
		else
		{
			if(!arg.paramInOut.equals("in"))
				pw.println(arg.paramName + ".value = java.lang.Float.intBitsToFloat(fval);");
			else
				pw.println(arg.paramName + " = java.lang.Float.intBitsToFloat(fval);");
		}
		return index;		
	}

	/*********************************************************
	* Prints the un-marshaling of double
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethodParam The method parameter
	*
	* @param int The current index
	*	
	* @return int The new index
	*********************************************************/
	public int doUnMarshalDouble(MQCPrinter pw, InterfaceMethodParam arg, int index)
	{
		if(!doubleDef)
			pw.println("long dval = 0;");
		if(!longDef)
		{
			pw.println("long lval1 = 0;");
			pw.println("long lval2 = 0;");
		}
		longDef = true;
		doubleDef = true;		
		int tempIndex;
		
		int paramSize = codeGenHelper.getParamSize("double");

		pw.println("if(bigEndian)");
		pw.println("{");
		pw.setIndent(pw.getIndent()+1);

		tempIndex = index;
		pw.print("lval1 = (long) (" );
		for(int i = (paramSize/2)-1; i > 0; i--)
		{
			pw.nprint("((_msg[" + index++ + "] & 0xFF) << " + (8 * i) + ") | ");
		}
		pw.nprintln("(_msg[" + index++ + "] & 0xFF));");
		pw.print("lval2 = (long) (" );
		for(int i = (paramSize/2)-1; i > 0; i--)
		{
			pw.nprint("((_msg[" + index++ + "] & 0xFF) << " + (8 * i) + ") | ");
		}
		pw.nprintln("(_msg[" + index++ + "] & 0xFF));");
		pw.println("dval = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");

		pw.setIndent(pw.getIndent()-1);
		pw.println("}");
		pw.println("else");
		pw.println("{");
		pw.setIndent(pw.getIndent()+1);
		pw.print("lval1 = (long) (" );
		for(int i = 0; i < (paramSize/2)-1; i++)
		{
			pw.nprint("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * i) + ") | ");
		}
		pw.nprintln("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * ((paramSize/2)-1)) + "));");
		pw.print("lval2 = (long) (" );
		for(int i = 0; i < (paramSize/2)-1; i++)
		{
			pw.nprint("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * i) + ") | ");
		}
		pw.nprintln("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * ((paramSize/2)-1)) + "));");
		pw.println("dval = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");
		
		pw.setIndent(pw.getIndent()-1);
		pw.println("}");

		if(arg == null)
		{
			pw.println("_result = java.lang.Double.longBitsToDouble(dval);");
		}
		else
		{
			if(!arg.paramInOut.equals("in"))
				pw.println(arg.paramName + ".value = java.lang.Double.longBitsToDouble(dval);");
			else
				pw.println(arg.paramName + " = java.lang.Double.longBitsToDouble(dval);");
		}
		return index;		
	}

	/*********************************************************
	* Prints the un-marshaling of long in a client
	* 
	* @param MQCPrinter The outputstream.
	*        
	* @param InterfaceMethodParam The method parameter
	*
	* @param int The current index
	*	
	* @return int The new index
	*********************************************************/
	public int doUnMarshalLong(MQCPrinter pw, InterfaceMethodParam arg, int index)
	{
		if(!longDef)
		{
			pw.println("long lval1 = 0;");
			pw.println("long lval2 = 0;");
		}
		longDef = true;
		int tempIndex;
		
		int paramSize = codeGenHelper.getParamSize("long long");
		pw.println("if(bigEndian)");
		pw.println("{");
		pw.setIndent(pw.getIndent()+1);

		pw.print("lval1 = (long) (" );

		tempIndex = index;
		for(int i = (paramSize/2)-1; i > 0; i--)
		{
			pw.nprint("((_msg[" + index++ + "] & 0xFF) << " + (8 * i) + ") | ");
		}
		pw.nprintln("(_msg[" + index++ + "] & 0xFF));");
		pw.print("lval2 = (long) (" );
		for(int i = (paramSize/2)-1; i > 0; i--)
		{
			pw.nprint("((_msg[" + index++ + "] & 0xFF) << " + (8 * i) + ") | ");
		}
		pw.nprintln("(_msg[" + index++ + "] & 0xFF));");
		if(arg == null)
		{
			pw.println("_result = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");
		}
		else
		{
			if(!arg.paramInOut.equals("in"))
				pw.println(arg.paramName + ".value = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");
			else
				pw.println(arg.paramName + " = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));");
		}

		pw.setIndent(pw.getIndent()-1);
		pw.println("}");
		pw.println("else");
		pw.println("{");
		pw.setIndent(pw.getIndent()+1);
		pw.print("lval1 = (long) (" );
		for(int i = 0; i < (paramSize/2)-1; i++)
		{
			pw.nprint("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * i) + ") | ");
		}
		pw.nprintln("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * ((paramSize/2)-1)) + "));");
		pw.print("lval2 = (long) (" );
		for(int i = 0; i < (paramSize/2)-1; i++)
		{
			pw.nprint("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * i) + ") | ");
		}
		pw.nprintln("((_msg[" + tempIndex++ + "] & 0xFF) << " + (8 * ((paramSize/2)-1)) + "));");

		if(arg == null)
		{
			pw.println("_result = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");
		}
		else
		{
			if(!arg.paramInOut.equals("in"))
				pw.println(arg.paramName + ".value = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");
			else
				pw.println(arg.paramName + " = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));");
		}
		
		pw.setIndent(pw.getIndent()-1);
		pw.println("}");
		return index;
	}

}
