/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

include(`Config_macros.m4')

package mqc;

import mqc.ior.*;
import mqc.transports.*;
import mqc.holders.*;
import mqc.protocols.*;
import java.util.Vector;

/**
 * Base class for delegates.
 * 
 * @author Olav Haugan
 * @author Kim Christian Swenson
 *
 */
public class Delegate
{
   public static int GIOP_PROT = 0;
   public static int GIOP_LITE_PROT = 1;
   public static int MQC_IOP_PROT = 2;

   private int intInReqID;

   public ORB orb;
   public IOR ior;
   public int protocol;

   protected Transport transport;
   protected Vector vTransports;

   protected boolean bigEnd;

   protected Request m_request;
   protected Reply m_reply;

   __TEMPSPACE__(
   private Vector vMessageBuffer = new Vector();
   )


   /**
    * Constructor. Initializes variables.
    *
    * @param ORB The orb reference
    * @param IOR The reference to the ior.
    */
   public Delegate(ORB _orb, IOR _ior, boolean _big )
   {
      orb = _orb;
      ior = _ior;
      vTransports = new Vector();
      bigEnd = _big;
   }      

   /**
    * Creates an request message using CDRStream.
    *
    * @param String The name of the operation requested.
    * @param Boolean true if oneway call, false otherwise.
    * @param int Size of buffer needed.
    * @return CDROutputStream The output stream.
    */
   public CDROutputStream _request(String operation,
                                   boolean twoWay,
                                   int sizeOfBuffer)
   {
      __SPATIAL__(
      // here the request id is a class variable, so we can grab any of the
      //    transports to generate the id--there will always be at least one
      Transport tCurrent = (Transport) vTransports.elementAt(0);
      int reqID = tCurrent.getNextRequestID();
      )
      __NOSPATIAL__(
      int reqID = transport.getNextRequestID();
      )
      intInReqID = reqID;

      m_request.setRequestID( reqID );
      m_request.setOneWay( ! twoWay );
      m_request.setOperation( operation );
      m_request.setObjectKey( ior.getObjectKey() );
      m_request.setBigEndian( bigEnd );

      m_request.getBody().setMinSize( sizeOfBuffer );

      return new CDROutputStream( m_request.getBody().getBuffer(), 0, bigEnd);
   }

   /**
    * Creates an request message using a byte buffer.
    * Side: Client (Client->_xxxStub->ObjImpl->this)
    * @param String The name of the operation requested.
    * @param Boolean true if oneway call, false otherwise.
    * @param int Size of data.
    */
   public byte[] _request(String operation,
                          int sizeOfBuffer,
                          boolean twoWay,
                          IntHolder ptr)
   {
      __SPATIAL__(
      // here the request id is a class variable, so we can grab any of the
      //    transports to generate the id--there will always be at least one
      Transport tCurrent = (Transport) vTransports.elementAt(0);
      int reqID = tCurrent.getNextRequestID();
      )
      __NOSPATIAL__(
      int reqID = transport.getNextRequestID();
      )
      intInReqID = reqID;

      m_request.setRequestID( reqID );
      m_request.setOneWay( ! twoWay );
      m_request.setOperation( operation );
      m_request.setObjectKey( ior.getObjectKey() );
      m_request.setBigEndian( bigEnd );

      // Have Stub code write from beginning of buffer
      ptr.value = 0;

      m_request.getBody().setMinSize( sizeOfBuffer );
      return m_request.getBody().getBuffer();
   }

   /**
    * @param out  output stream containing already marshaled message body
    * @return  input stream containing message body ready to be unmarshaled
    */
   public CDRInputStream _invoke( CDROutputStream out )
   {
      byte[] buf = _invoke( out.getBuffer(), new IntHolder( out.bytesStreamed() ), out.isBigEndian() );
      return new CDRInputStream( buf, 0, m_reply.isBigEndian() );
   }

   /**
    * @param buffer  pre-marshalled message body
    * @param ptr  value contains number of bytes in body
    * @param toWay  this parameter does not matter, as whether
    *               the message is twoWay or not was already 
    *               determined when request() was called
    * @return  buffer containing message body
    */
   public byte[] _invoke(byte[] buffer, IntHolder ptr, boolean twoWay)
   {
      // Prepare message body for marshaling
      m_request.getBody().setBuffer( buffer );
      m_request.getBody().setPtr( ptr.value );

      __SPATIAL__(
      // here we can send for each transport--I should probably add code to make this happen
      //    in separate threads??
      //Make several invokation if we use Spatial Redundancy
      for(int i=0;i<vTransports.size();i++)
      {
         Transport tCurrent = (Transport) vTransports.elementAt(i);
         __DEBUG__( System.out.println("Sending a msg for transport: " + i);)
         tCurrent.send( m_request );
      }
      )
      __NOSPATIAL__(
      //Make the invokation to the server
      transport.send( m_request );
      )

      if( ! m_request.isOneWay() )
      {
         __SPATIAL__(
         for(int i=0;i<vTransports.size();i++)
         {
            Transport tCurrent = (Transport) vTransports.elementAt(i);
            __DEBUG__( System.out.println("Waiting for a response from transport: " + i);)
            while( ! tCurrent.waitForReply( m_reply ) );
         }
         )

         boolean gotReply = false;
         __TEMPORAL__(
         // set up a loop to keep looking for the reply
         int count = 0;
         while ( count++ < Config.temporal && ! gotReply )
         {
            gotReply = transport.waitForReply( m_reply );
         }
         if ( gotReply )
         {
            // KCS: m_reply is passed for now, if variable size 
            // is to be implemented later, a clone needs to be
            // passed as argument here.
            __DEBUG5__(System.err.println( "Delegate.invoke(): gotReply = true, skipping duplicates. . ." );)
            transport.skipReply( m_reply );
            __DEBUG5__(System.err.println( "Delegate.invoke(): done." );)
         }
         )
         __NOTEMPSPACE__(
         /*
          * Expecting a Reply message, 
          * Transport demarshals the header
          */
         gotReply = transport.waitForReply( m_reply );
         )

         if ( ! gotReply )
         {
            throw new RuntimeException( "Did not get a valid reply message." );
         }

         switch ( m_reply.getReplyStatus() )
         {
         case Reply.NO_EXCEPTION:
            break;
         case Reply.USER_EXCEPTION:
            throw new mqc.exception.UserException( m_reply.getExceptionID() );
         case Reply.SYSTEM_EXCEPTION:
            throw new mqc.exception.SystemException( m_reply.getExceptionID(),
                                                     m_reply.getExceptionMinorCode(),
                                                     m_reply.getCompletionStatus() );
         default:
            System.out.println("Don't Understand Reply Message " + m_reply.getReplyStatus());
         }
      }

      // Message body starts at index 0 of body buffer
      ptr.value = 0;

      return m_reply.getBody().getBuffer();
   }

   /**
    * Sets the transport.
    *
    * @param Transport The transport object to be assigned
    *
    */
   public void set_transport(Transport transport)
   {
      this.transport = transport;
      vTransports.addElement(transport);
   }

   /**
    * Adds a transport to the transport vector.
    *
    * @param Transport The transport object to be assigned
    *
    */
   public void add_transport(Transport transport)
   {
      if(vTransports.size() == 0)
      {
         this.transport = transport;
      }
      vTransports.addElement(transport);
   }

   /**
    * Gets the transport
    *
    * @return Transport The transport
    *
    */
   public Transport get_transport()
   {
      Transport tCurrent = (Transport) vTransports.elementAt(0);
      return tCurrent;
      //return transport;
   }

   /**
    * Gets the vector of transports.
    *
    * @param Transport The transport object to be assigned
    *
    */
   public Vector get_transports()
   {
      return vTransports;
   }

   /**
    * Gets the endianess
    *
    * @return boolean true if big endian, false otherwise
    *
    */
   public boolean isBigEndian()
   {
      return bigEnd;
   }
}
