/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

include(`Config_macros.m4')

package mqc.ior;

import mqc.holders.*;
import java.util.*;

/**
* IOR parsers. Converts an IOR string to readable and
* usable parameters
*
* @author Olav Haugan  - Initially
* @author A. David McKinnon - Initially
* @author Eivind Naess - UDPProfiles/UDPMacros/corbaloc
*/
public class IORParser
{
   /* TaggedProfile members */
   private Profile profile;

   /**
    * Constructor. Initializes variables
    */
   public IORParser()
   {
      profile = null;
   }

   /**
    * Parses the IOR string
    *
    * @param String The IOR string
    *
    * @return boolean true if string is valid, false otherwise
    */
   public boolean parseIOR(String ior)
   {
      profile = new Profile();

      boolean bigEndian = true;
      if(!ior.substring(0,4).equals("IOR:"))
      {
         System.out.println("Unknown IOR: format: " + ior.substring(0,3));
         return false;
      }

      int index = 0;
      int len = ior.length()-4;

      byte buf[] = new byte[len/2];
      int ptr = 0;

      /* Convert string to hex numbers */
      for(index = 0; index < len; index += 2)
      {
         buf[ptr] = deHexify(ior.substring(4+index, 6+index));
         ptr++;
      }
      ptr = 0;
      /* Big endian or little endian? */
      if(buf[ptr] == 1)
      {
         bigEndian = false;
      }
      ptr++;

      //align to 4 byte value
      ptr = (ptr + 3) & ~3;

      int typeIDLen = 0;
      if(bigEndian)
      {
         typeIDLen = ((buf[ptr++] & 0xFF) << 24) |
                     ((buf[ptr++] & 0xFF) << 16) |
                     ((buf[ptr++] & 0xFF) << 8) |
                     ((buf[ptr++] & 0xFF));
      }
      else
      {
         typeIDLen = (buf[ptr++] & 0xFF) |
                     ((buf[ptr++] & 0xFF) << 8) |
                     ((buf[ptr++] & 0xFF) << 16) |
                     ((buf[ptr++] & 0xFF) << 24);
      }

      StringBuffer tmp = new StringBuffer();
      /* Get the type id */
      int i = 0;
      for(i = 0; i < typeIDLen-1; i++)
      {
         tmp.append( (char) buf[ptr]);
         ptr++;
      }
      ptr++;

      profile.typeID = tmp.toString();

      //align to 4 byte value
      ptr = (ptr + 3) & ~3;
      int noProfiles = 0;
      if(bigEndian)
      {
         noProfiles = ((buf[ptr++] & 0xFF) << 24) |
                      ((buf[ptr++] & 0xFF) << 16) |
                      ((buf[ptr++] & 0xFF) << 8) |
                      (buf[ptr++] & 0xFF);
      }
      else
      {
         noProfiles = (buf[ptr++] & 0xFF) |
                      ((buf[ptr++] & 0xFF) << 8) |
                      ((buf[ptr++] & 0xFF) << 16) |
                      ((buf[ptr++] & 0xFF) << 24);
      }

      /* Parse each profile in IOR */
      for(i = 0; i < noProfiles; i++)
      {
         ptr = parseTaggedProfile(buf, ptr, bigEndian);
         if(ptr == -1)
            return false;
      }
      return true;
   }

   /**
    * Parses the corbaloc string
    *
    * @param String The corbaloc string
    *
    * @return boolean true if string is valid, false otherwise
    */
   public boolean parseCorbaloc(String corbaloc)
   {
      profile = new Profile();

      boolean bigEndian = true;
      int     major     = 1;
      int     minor     = 1;

      if(!corbaloc.substring(0,9).equals("corbaloc:"))
      {
         __DEBUG__(System.out.println("Unknown corbaloc: " + corbaloc.substring(0,9));)
         return false;
      }

      int ptr = 9;
      TaggedProfile taggedProfile = null;
      __DEBUG__(System.out.println(corbaloc);)

      //Defaualt corbaloc::... => = use IIOPProfile!
      if (corbaloc.substring(ptr, ptr+1).equals(":"))
      {
         taggedProfile = new IIOPProfile();
         ptr += 1;
      }
      __TCP_TP__( __GIOP_PROTOCOL__(
      if (corbaloc.substring(ptr, ptr+5).equals("iiop:"))
      {
         taggedProfile = new IIOPProfile();
         ptr += 5;
      }
      ))
      __TCP_TP__( __GIOPLITE_PROTOCOL__(
      if (corbaloc.substring(ptr, ptr+9).equals("iioplite:"))
      {
         taggedProfile = new IIOPLiteProfile();
         ptr += 9;
      }
      ))
      __TCP_TP__( __MQCIOP_PROTOCOL__(
      if (corbaloc.substring(ptr, ptr+7).equals("mqciop:"))
      {
         taggedProfile = new MQCIOPProfile();
         ptr += 7;
      }
      ))

      __UDP_TP__( __GIOP_PROTOCOL__(
      if (corbaloc.substring(ptr, ptr+5).equals("uiop:"))
      {
         taggedProfile = new UIOPProfile();
         ptr += 5;
      }
      ))
      __UDP_TP__( __GIOPLITE_PROTOCOL__(
      if (corbaloc.substring(ptr, ptr+9).equals("uioplite:"))
      {
         taggedProfile = new UIOPLiteProfile();
         ptr += 9;
      }
      ))
      __UDP_TP__( __MQCIOP_PROTOCOL__(
      if (corbaloc.substring(ptr, ptr+8).equals("umqciop:"))
      {
         taggedProfile = new UMQCIOPProfile();
         ptr += 8;
      }
      ))

      if (taggedProfile == null)
      {
         //This should never happen!
         __DEBUG__(System.out.println("**Warning: IORParser assumes IIOPProfile, please check your settings\n" + corbaloc);)
         return false;
         //taggedProfile = new IIOPProfile();
         //ptr += corbaloc.substring(ptr, corbaloc.length()).indexOf(":"); //justify ptr
      }
      ptr = taggedProfile.parseCorbaloc(corbaloc.substring(ptr,corbaloc.length()));
      profile.add(taggedProfile);

      return true;
   }

   /**
    * Converts a hex string to its byte value
    *
    * @param String The hex string
    *
    * @return byte The byte value of the hex string
    */
   private byte deHexify(String value)
   {
      return (byte) Integer.parseInt(value, 16);
   }

   /**
    * Converts a byte to an hex string 
    *
    * @param byte The byte value of the hex string
    *
    * @return String The hex string
    */
   private String hexify(byte value)
   {
      return (String) (hex((value & 0xF0) >> 4) + hex(value & 0x0F));
   }

   /**
    * Converts a byte to an hex string 
    *
    * @param byte The byte value of the hex string
    *
    * @return String The hex string
    */
   private String hex(int value)
   {
      switch(value)
      {
      case 10:
         return "A";
      case 11:
         return "B";
      case 12:
         return "C";
      case 13:
         return "D";
      case 14:
         return "E";
      case 15:
         return "F";
      default:
         return String.valueOf(value);
      }
   }

   /**
    * Parses a tagged profile. 
    *
    * @param byte[] The buffer with the IOR information to be parsed.
    * @param int The pointer in the buffer
    * @param boolean true if big endian, false otherwise
    * @return int The updated pointer position.
    */
   private int parseTaggedProfile(byte[] buf, int ptr, boolean bigEndian)
   {
      //align to 4 byte value
      ptr = (ptr + 3) & ~3;

      //read in profile tag
      int profileTag = 0;

      if(bigEndian)
      {
         profileTag = ((buf[ptr++] & 0xFF) << 24) |
                      ((buf[ptr++] & 0xFF) << 16) |
                      ((buf[ptr++] & 0xFF) << 8) |
                      (buf[ptr++] & 0xFF);
      }
      else
      {
         profileTag = (buf[ptr++] & 0xFF) |
                      ((buf[ptr++] & 0xFF) << 8) |
                      ((buf[ptr++] & 0xFF) << 16) |
                      ((buf[ptr++] & 0xFF) << 24);
      }

      //align to 4 byte value
      ptr = (ptr + 3) & ~3;

      /* Get the length of the profile sequence */
      int seqLen = 0;
      if(bigEndian)
      {
         seqLen = ((buf[ptr++] & 0xFF) << 24) |
                  ((buf[ptr++] & 0xFF) << 16) |
                  ((buf[ptr++] & 0xFF) << 8) |
                  (buf[ptr++] & 0xFF);
      }
      else
      {
         seqLen = (buf[ptr++] & 0xFF) |
                  ((buf[ptr++] & 0xFF) << 8) |
                  ((buf[ptr++] & 0xFF) << 16) |
                  ((buf[ptr++] & 0xFF) << 24);
      }

      TaggedProfile taggedProfile = null;
      switch(profileTag)
      {
      __TCP_TP__(
      __GIOPLITE_PROTOCOL__(
      case TaggedProfile.IIOPLITE:
         taggedProfile = new IIOPLiteProfile();
         break;
      )
      __MQCIOP_PROTOCOL__(
      case TaggedProfile.MQCIOP:
         taggedProfile = new MQCIOPProfile();
         break;
      )
      __GIOP_PROTOCOL__(
      case TaggedProfile.IIOP:
         taggedProfile = new IIOPProfile();
         break;
      )
      )
      __UDP_TP__(
      __GIOPLITE_PROTOCOL__(
      case TaggedProfile.UIOPLITE:
         taggedProfile = new UIOPLiteProfile();
         break;
      )
      __MQCIOP_PROTOCOL__(
      case TaggedProfile.UMQCIOP:
         taggedProfile = new UMQCIOPProfile();
         break;
      )
      __GIOP_PROTOCOL__(
      case TaggedProfile.UIOP:
         taggedProfile = new UIOPProfile();
         break;
      )
      )
      default:
      taggedProfile = null;
         break;
      }
      if ( taggedProfile != null )
      {
         ptr = taggedProfile.parse(buf, ptr, seqLen);
         profile.add(taggedProfile);
      }
      else
      {
         throw new RuntimeException("IOR TAG NOT IMPLEMENTED");
      }
      return ptr;
   }

   /**
    * Returns a profile
    *
    * @return Profile The parsed profile
    */
   public Profile getProfile()
   {
      return profile;
   }

   /**
    * Encodes a profile to an IOR string
    *
    * @param Profile The profile
    *
    * @return String The IOR string
    */
   public String encodeProfile(Profile profile)
   {
      this.profile = profile;

      //big endian byte + padding + sizeOfID + ID + `''\0`'';
      int size = 1 + 3 + 4 + profile.typeID.length() + 1;
      //align to 4 byte value
      size = (size + 3) & ~3;
      //Number of profiles
      size += 4;

      Enumeration e = profile.taggedProfile.elements();
      TaggedProfile tagProfile = null;

      while(e.hasMoreElements())
      {
         tagProfile = (TaggedProfile) e.nextElement();
         size = tagProfile.size(size);
      }

      byte ior[] = new byte[size];
      for(int i = 0; i < size; i++)
         ior[i] = (byte) 0;

      int ptr = 0;

      //big endian
      ior[ptr++] = (byte) 0;

      //padding
      ior[ptr++] = (byte) 0;
      ior[ptr++] = (byte) 0;
      ior[ptr++] = (byte) 0;

      //type id length
      ior[ptr++] = (byte) ((profile.typeID.length() + 1) >> 24);
      ior[ptr++] = (byte) ((profile.typeID.length() + 1) >> 16);
      ior[ptr++] = (byte) ((profile.typeID.length() + 1) >> 8);
      ior[ptr++] = (byte) (profile.typeID.length() + 1);

      //type id
      byte[] typeID = profile.typeID.getBytes();
      System.arraycopy(typeID, 0, ior, ptr, typeID.length);
      ptr += typeID.length;
      ior[ptr++] = (byte) 0;

      //align to 4 byte value
      ptr = (ptr + 3) & ~3;

      //Number of profiles
      ior[ptr++] = (byte) (profile.taggedProfile.size() >> 24);
      ior[ptr++] = (byte) (profile.taggedProfile.size() >> 16);
      ior[ptr++] = (byte) (profile.taggedProfile.size() >> 8);
      ior[ptr++] = (byte) profile.taggedProfile.size();

      IntHolder ptrHolder = new IntHolder();
      ptrHolder.value = ptr;

      e = profile.taggedProfile.elements();
      while(e.hasMoreElements())
      {
         tagProfile = (TaggedProfile) e.nextElement();
         ior = tagProfile.encodeProfile(ior, ptrHolder);
      }

      StringBuffer iorStr = new StringBuffer("IOR:");
      for(int i = 0; i < ior.length; i++)
      {
         iorStr.append(hexify(ior[i]));
      }
      return iorStr.toString();
   }

   /**
    * Encodes a profile to a corbaloc string.
    *   At this time only the first profile is encoded.
    *
    * @param   Profile The profile
    * @return String  The corbaloc string
    */
   public String encodeCorbalocProfile(Profile profile)
   {
      this.profile = profile;

      String corbalocStr = new String("corbaloc:");

      Enumeration e = profile.taggedProfile.elements();
      TaggedProfile tagProfile = null;
      int j = profile.taggedProfile.size();
      for (int i=0; (i < j) && (e.hasMoreElements() == true); i++)
      {
         tagProfile = (TaggedProfile) e.nextElement();
         corbalocStr += tagProfile.encodeCorbalocProfile("");
         if (i < j-1 && j != 1)
         {
            corbalocStr += ",";
         }
      }

      return corbalocStr.toString();
   }
}

