/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.protocols;

/**
 * An Interface represening any request message of either MQC|GIOP|GIOPLite.
 *
 * @author Kim Christian Swenson
 */
public class GIOPRequest implements GIOPMessage, Request
{
   protected GIOPHeader m_GIOPHeader;
   protected GIOPRequestHeader m_GIOPRequestHeader;
   protected Body m_body;

   public GIOPRequest( GIOPHeader header, GIOPRequestHeader reqHeader )
   {
      m_GIOPHeader = header;
      m_GIOPHeader.setMessageType( GIOPMessage.REQUEST_MSG );
      m_GIOPRequestHeader = reqHeader;
      m_body = new Body();
   }

   public GIOPHeader getGIOPHeader()
   {
      return m_GIOPHeader;
   }

   public void setGIOPHeader( GIOPHeader header )
   {
      m_GIOPHeader = header;
   }

   public GIOPRequestHeader getGIOPRequestHeader()
   {
      return m_GIOPRequestHeader;
   }

   public void setGIOPRequestHeader( GIOPRequestHeader header )
   {
      m_GIOPRequestHeader = header;
   }

   public Body getBody()
   {
      return m_body;
   }

   public void setBody( Body body )
   {
      m_body = body;
   }

   public void setBigEndian( boolean bigEndian )
   {
      m_GIOPHeader.setBigEndian( bigEndian );
   }

   public boolean isBigEndian()
   {
      return m_GIOPHeader.isBigEndian();
   }

   public void deMarshal( CDRInputStream in ) throws RuntimeException
   {
      // Reset stream
      in.reset();

      // GIOP header
      m_GIOPHeader.deMarshal( in );

      // GIOP Request header
      m_GIOPRequestHeader.deMarshal( in );

      // GIOP Request Body

      // Copy data from underlying buffer of 
      // CDRInputStream into body buffer, and adjust
      // the Stream's pointer acordingly.

      // Get bufferPointer and align it with 4-bytes
      int bufPtr = in.bytesStreamed();
      bufPtr = (bufPtr + 3) & ~3;

      // Make sure message body's buffer is big enough
      m_body.setMinSize( m_GIOPHeader.getMessageSize() - bufPtr + getHdrSize() );

      // copy body part of message from inputstream into body
      System.arraycopy( in.getBuffer(), bufPtr, m_body.getBuffer(), 0, m_GIOPHeader.getMessageSize() - bufPtr + getHdrSize() );

      // Adjust buffer pointer
      in.setPtr( m_GIOPHeader.getMessageSize() + getHdrSize() );
   }

   public void marshal( CDROutputStream out )
   {
      // Reset stream
      out.reset();

      // GIOP header
      m_GIOPHeader.marshal( out );

      // GIOP Request header
      m_GIOPRequestHeader.marshal( out );

      // GIOP Request Body

      // Copy data from body buffer into underlying
      // buffer of CDROutputStream, and adjust
      // the Stream's pointer acordingly.

      // get underlying buffer
      byte[] buf = out.getBuffer();
      
      // Get bufferPointer and align it with 4-bytes
      int bufPtr = out.bytesStreamed();
      bufPtr = (bufPtr + 3) & ~3;

      // Hack ... set message size again now that we know it .. !
      m_GIOPHeader.setMessageSize( bufPtr - getHdrSize() + m_body.getPtr() );
      GIOPHeader.setMsgSize( buf, m_GIOPHeader.getMessageSize(), m_GIOPHeader.isBigEndian() );

      // Resize if necessary
      if ( buf.length < bufPtr + m_body.getPtr() )
      {
         byte[] tmp = new byte[bufPtr + m_body.getPtr()];
         System.arraycopy(buf, 0, tmp, 0, out.bytesStreamed() );
         buf = tmp;
      }

      // copy data
      System.arraycopy( m_body.getBuffer(), 0, buf, bufPtr, m_body.getPtr() );

      // Insert underlying buffer in case of resize
      out.setBuffer( buf );

      // Adjust buffer pointer
      out.setPtr( bufPtr + m_body.getPtr() );
   }

   public int getRequestID()
   {
      return m_GIOPRequestHeader.getRequestID();
   }

   public void setRequestID( int id )
   {
      m_GIOPRequestHeader.setRequestID( id );
   }

   public boolean isOneWay()
   {
      return m_GIOPRequestHeader.isOneWay();
   }

   public void setOneWay( boolean oneway )
   {
      m_GIOPRequestHeader.setOneWay( oneway );
   }

   public byte[] getObjectKey()
   {
      return m_GIOPRequestHeader.getObjectKey();
   }

   public void setObjectKey( byte[] key )
   {
      m_GIOPRequestHeader.setObjectKey( key );
   }

   public String getOperation()
   {
      return m_GIOPRequestHeader.getOperation();
   }

   public void setOperation( String op )
   {
      m_GIOPRequestHeader.setOperation( op );
   }

   public void setMsgSize( byte[] buf, int size, boolean bigEndian )
   {
      GIOPHeader.setMsgSize( buf, size, bigEndian );
   }

   public int getHdrSize()
   {
      return 12;
   }

}

