/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.protocols;

/**
 * MQC-IOP request message version 1.0 protocol message.
 * 
 * struct MQCIOPRequestHeader {
 *    unsigned long     request_id;
 *    octet             response_expected;
 *    octet             object_key[15];
 *    string            operation;
 * };
 * 
 * @author Kim Christian Swenson
 *
 */
public class MQCIOPRequest implements Request
{
   protected MQCIOPHeader m_MQCIOPHeader;

   protected int m_reqID;

   protected boolean m_oneWay;

   protected byte[] m_objectKey;

   protected String m_operation;
   
   protected Body m_body;

   public MQCIOPRequest( )
   {
      this( new MQCIOPHeader() );
   }

   public MQCIOPRequest( MQCIOPHeader header )
   {
      m_MQCIOPHeader = header;
      header.setType( MQCIOPHeader.REQUEST_MSG );
      m_body = new Body();
      m_objectKey = new byte[15];
   }

   public Body getBody()
   {
      return m_body;
   }

   public void setBody( Body body )
   {
      m_body = body;
   }

   public int getRequestID()
   {
      return m_reqID;
   }

   public void setRequestID( int id )
   {
      m_reqID = id;
   }

   public boolean isOneWay()
   {
      return m_oneWay;
   }

   public void setOneWay( boolean oneway )
   {
      m_oneWay = oneway;
   }

   public byte[] getObjectKey()
   {
      return m_objectKey;
   }

   public void setObjectKey( byte[] key )
   {
      m_objectKey = key;
   }

   public String getOperation()
   {
      return m_operation;
   }

   public void setOperation( String op )
   {
      m_operation = op;
   }

   public void setMsgSize( byte[] buf, int value, boolean bigEndian )
   {
      MQCIOPHeader.setMsgSize( buf, value, bigEndian );
   }

   public int getHdrSize()
   {
      return 12;
   }

   public boolean isBigEndian()
   {
      return m_MQCIOPHeader.isBigEndian();
   }

   public void setBigEndian( boolean bigEndian )
   {
      m_MQCIOPHeader.setBigEndian( bigEndian );
   }

   public void marshal( CDROutputStream out )
   {
      out.reset();

      boolean origAlign = out.isAlign();
      out.setAlign( false );

      // MQCIOP header
      m_MQCIOPHeader.marshal( out );

      // unsigned long request_id
      out.writeUnsignedLong( m_reqID );

      // octet response_expected
      out.writeOctet( m_oneWay ? (byte)0x00 : (byte)0x01 );

      // octet object_key[15]
      out.writeOctet( m_objectKey[0] );
      out.writeOctet( m_objectKey[1] );
      out.writeOctet( m_objectKey[2] );
      out.writeOctet( m_objectKey[3] );
      out.writeOctet( m_objectKey[4] );
      out.writeOctet( m_objectKey[5] );
      out.writeOctet( m_objectKey[6] );
      out.writeOctet( m_objectKey[7] );
      out.writeOctet( m_objectKey[8] );
      out.writeOctet( m_objectKey[9] );
      out.writeOctet( m_objectKey[10] );
      out.writeOctet( m_objectKey[11] );
      out.writeOctet( m_objectKey[12] );
      out.writeOctet( m_objectKey[13] );
      out.writeOctet( m_objectKey[14] );

      // string operation
      out.writeString( m_operation );

      // MQCIOPRequest Body

      // Copy data from body buffer into underlying
      // buffer of CDROutputStream, and adjust
      // the Stream's pointer acordingly.

      // get underlying buffer
      byte[] buf = out.getBuffer();
      
      // Get bufferPointer and align it with 4-bytes
      int bufPtr = out.bytesStreamed();
      bufPtr = (bufPtr + 3) & ~3;

      // Hack ... set message size again now that we know it .. !
      m_MQCIOPHeader.setSize( bufPtr - getHdrSize() + m_body.getPtr() );
      MQCIOPHeader.setMsgSize( buf, m_MQCIOPHeader.getSize(), m_MQCIOPHeader.isBigEndian() );

      // Resize if necessary
      if ( buf.length < bufPtr + m_body.getPtr() )
      {
         byte[] tmp = new byte[bufPtr + m_body.getPtr()];
         System.arraycopy(buf, 0, tmp, 0, out.bytesStreamed() );
         buf = tmp;
      }

      // copy data
      System.arraycopy( m_body.getBuffer(), 0, buf, bufPtr, m_body.getPtr() );

      // Insert underlying buffer in case of resize
      out.setBuffer( buf );

      // Adjust buffer pointer
      out.setPtr( bufPtr + m_body.getPtr() );

      out.setAlign( origAlign );
   }

   public void deMarshal( CDRInputStream in ) throws RuntimeException
   {
      in.reset();

      boolean origAlign = in.isAlign();
      in.setAlign( false );

      // MQCIOP header
      m_MQCIOPHeader.deMarshal( in );

      // unsigned long request_id
      m_reqID = in.readUnsignedLong( );

      // octet response_expected
      m_oneWay = (in.readOctet() == (byte)0x00);

      // octet object_key[15]
      m_objectKey[0] = in.readOctet( );
      m_objectKey[1] = in.readOctet( );
      m_objectKey[2] = in.readOctet( );
      m_objectKey[3] = in.readOctet( );
      m_objectKey[4] = in.readOctet( );
      m_objectKey[5] = in.readOctet( );
      m_objectKey[6] = in.readOctet( );
      m_objectKey[7] = in.readOctet( );
      m_objectKey[8] = in.readOctet( );
      m_objectKey[9] = in.readOctet( );
      m_objectKey[10] = in.readOctet( );
      m_objectKey[11] = in.readOctet( );
      m_objectKey[12] = in.readOctet( );
      m_objectKey[13] = in.readOctet( );
      m_objectKey[14] = in.readOctet( );

      // string operation
      m_operation = in.readString( );

      // MQCIOPRequest Body

      // Copy data from underlying buffer of 
      // CDRInputStream into body buffer, and adjust
      // the Stream's pointer acordingly.

      // Get bufferPointer and align it with 4-bytes
      int bufPtr = in.bytesStreamed();
      bufPtr = (bufPtr + 3) & ~3;

      // Make sure message body's buffer is big enough
      m_body.setMinSize( m_MQCIOPHeader.getSize() - bufPtr + getHdrSize() );

      // copy body part of message from inputstream into body
      System.arraycopy( in.getBuffer(), bufPtr, m_body.getBuffer(), 0, m_MQCIOPHeader.getSize() - bufPtr + getHdrSize() );

      // Adjust buffer pointer
      in.setPtr( m_MQCIOPHeader.getSize() + getHdrSize() );

      in.setAlign( origAlign );
   }

}
