/* Id: Mac.java,v 1.5 2000/01/15 03:25:04 gelderen Exp
 *
 * Copyright (C) 1995-1999 The Cryptix Foundation Limited.
 * All rights reserved.
 *
 * Use, modification, copying and distribution of this software is subject
 * the terms and conditions of the Cryptix General Licence. You should have
 * received a copy of the Cryptix General Licence along with this library;
 * if not, you can download a copy from http://www.cryptix.org/ .
 */
package mqc.security;


import mqc.security.InvalidKeyException;
import mqc.security.Key;
import mqc.security.NoSuchAlgorithmException;


/**
 * A MAC (Message Authentication Code).
 *
 * @author Jeroen C. van Gelderen (gelderen@cryptix.org)
 * @version Revision: 1.5 
 */
public abstract class Mac
{
   private static final int
      STATE_UNDEFINED   = 0,
      STATE_INITIALIZED = 1;

   private int state;

   //FIXME: include the algorithm name later (not supported by Cryptix impl.s)
   //FIXME: private final String   algorithm;

   protected Mac()
   {
      this(STATE_UNDEFINED);
   }


   /**
    * Construct a Mac with a given initial state.
    *
    * This constructor comes in handy when cloning.
    */
   private Mac(int state)
   {
      //this.algorithm = algorithm;
      this.state     = state;
   }

   //FIXME: Add support for the algortihm names
   //    public final String getAlgorithm() {
   //        return this.algorithm;
   //    }


   public final int getMacLength()
   {
      return engineGetMacLength();
   }
   
   public final void init(Key key)
      throws InvalidKeyException
   {

      engineInit(key);
      state = STATE_INITIALIZED;

      engineReset();  // just to be sure, clear out old stuff
   }


   public final void update(byte input)
      throws IllegalStateException
   {

      if(this.state != STATE_INITIALIZED)
      throw new IllegalStateException();

      engineUpdate(input);
   }


   public final void update(byte[] input)
      throws IllegalStateException
   {

      if(state != STATE_INITIALIZED)
         throw new IllegalStateException();

      engineUpdate(input, 0, input.length);
   }


   public final void update(byte[] input, int offset, int len)
      throws IllegalStateException
   {
      if(state != STATE_INITIALIZED)
         throw new IllegalStateException();

      //DEBUG        System.out.println("MAC::update(input, " + offset + ", " + len + ")");
      //DEBUG        for (int i = 0;  i < len;  i++ ) {
      //DEBUG            System.out.print(input[offset+i] + " " );
      //DEBUG        }
      //DEBUG        System.out.println();
      engineUpdate(input, offset, len);
   }


   public final byte[] doFinal()
      throws IllegalStateException
   {

      if (state != STATE_INITIALIZED)
         throw new IllegalStateException();

      byte[] res = engineDoFinal();
      reset();
      return res;
   }


   public final void doFinal(byte[] output, int outOffset)
      throws ShortBufferException, IllegalStateException
   {
      if (state != STATE_INITIALIZED)
         throw new IllegalStateException();

      int my_len = getMacLength();

      if ( (output.length-outOffset) < my_len )
         throw new ShortBufferException("Buffer too short");

      byte[] res = engineDoFinal();
      System.arraycopy(res, 0, output, outOffset, my_len);
      reset();
   }


   public final byte[] doFinal(byte[] input)
      throws IllegalStateException
   {
      if(state != STATE_INITIALIZED)
         throw new IllegalStateException();

      update(input);
      return doFinal();
   }

   public final void reset()
   {
      engineReset();
   }

   /**
    * Compares two MACs for equality. Does a simple byte compare.
    * 
    * @param maca one of the MACs to compare.
    * @param macb the other MAC to compare.    
    * @return true if the MACs are equal, false otherwise.
    */
   public static boolean isEqual(byte maca[], byte macb[])
   {
      if (maca.length != macb.length)
      {
         //DEBUG            System.out.println("Mismatched lengths: " + maca.length +
         //DEBUG                                " " + macb.length);
         return false;
      }

      for (int i = 0; i < macb.length; i++)
      {
         if (maca[i] != macb[i])
         {
            //DEBUG                System.out.println("Mismatch at char #:" + i + " " +
            //DEBUG                                maca[i] + " != " + macb[i]);
            //DEBUG                System.out.print("macA: ");
            //DEBUG                for (int j = 0;  j < maca.length;  j++) {
            //DEBUG                    System.out.print(maca[j] + " " );
            //DEBUG                }
            //DEBUG                System.out.print("\nmacB: ");
            //DEBUG                for (int j = 0;  j < macb.length;  j++) {
            //DEBUG                    System.out.print(macb[j] + " " );
            //DEBUG                }
            //DEBUG                System.out.println();
            return false;
         }
      }
      //DEBUG                System.out.print("\nmacA/B: ");
      //DEBUG                for (int j = 0;  j < macb.length;  j++) {
      //DEBUG                    System.out.print(macb[j] + " " );
      //DEBUG                }
      //DEBUG                System.out.println();
      return true;
   }


   // Mac SPI methods
   //...................
   protected abstract Object clone() throws CloneNotSupportedException;

   protected abstract int engineGetMacLength();

   protected abstract void engineInit(Key key) throws InvalidKeyException;

   protected abstract void engineUpdate(byte input);

   protected abstract void engineUpdate(byte[] input, int offset, int len);

   protected abstract byte[] engineDoFinal();

   protected abstract void engineReset();
}
